// Funge JIT compiler
// Copyright (c) 2004 Marek Materzok
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.

#include <cassert>

#include <llvm/Bytecode/Reader.h>
#include <llvm/ModuleProvider.h>
#include <llvm/Module.h>
#include <llvm/Type.h>
#include <llvm/DerivedTypes.h>
#include <llvm/Constants.h>
#include <llvm/Instructions.h>

#include "Bflib.hpp"

extern const unsigned char bflib_bytecode[];
extern const int bflib_bytecode_size;

namespace bflib
{

llvm::Function *main,
	*init,
	*begin_block,
	*end_block,
	*stack_transfer,
	*push_int,
	*pop_int,
	*push_int_many,
	*push_vector,
	*pop_vector,
	*input_int,
	*output_int,
	*input_char,
	*output_char,
	*fetch_char,
	*store_char,
	*fungespace_get,
	*fungespace_put,
	*go_away,
	*clear_stack;
static const llvm::Type *vector_v[] = {llvm::ArrayType::get(llvm::Type::IntTy, 3)};
const llvm::Type *vector = llvm::StructType::get(std::vector<const llvm::Type *>(vector_v, vector_v+1));
llvm::GlobalVariable *thread;

};

struct BflibFunction
{
	const char *name;
	llvm::Type *ret;
	llvm::Type **args;
	int nargs;
	bool vararg;
	llvm::Function **holder;
};

static llvm::Type *args_noargs[] = {};
static llvm::Type *args_1int[] = {llvm::Type::IntTy};
static llvm::Type *args_vector[] = {llvm::PointerType::get(bflib::vector)};
static llvm::Type *args_vector_int[] = {llvm::PointerType::get(bflib::vector), llvm::Type::IntTy};

BflibFunction bfuncs[] = 
{
	{"bfmain", llvm::Type::IntTy, args_noargs, 0, false, &bflib::main},
	{"bflib_init", llvm::Type::VoidTy, args_noargs, 0, false, &bflib::init},
	{"bflib_begin_block", llvm::Type::VoidTy, args_noargs, 0, false, &bflib::begin_block},
	{"bflib_end_block", llvm::Type::VoidTy, args_noargs, 0, false, &bflib::end_block},
	{"bflib_push_int", llvm::Type::VoidTy, args_1int, 1, false, &bflib::push_int},
	{"bflib_pop_int", llvm::Type::IntTy, args_noargs, 0, false, &bflib::pop_int},
	{"bflib_push_int_many", llvm::PointerType::get(llvm::Type::IntTy), args_1int, 1, false, &bflib::push_int_many},
	{"bflib_push_vector", llvm::Type::VoidTy, args_vector, 1, false, &bflib::push_vector},
	{"bflib_pop_vector", llvm::Type::VoidTy, args_vector, 1, false, &bflib::pop_vector},
	{"bflib_output_int", llvm::Type::VoidTy, args_1int, 1, false, &bflib::output_int},
	{"bflib_input_int", llvm::Type::IntTy, args_noargs, 0, false, &bflib::input_int},
	{"bflib_output_char", llvm::Type::VoidTy, args_1int, 1, false, &bflib::output_char},
	{"bflib_input_char", llvm::Type::IntTy, args_noargs, 0, false, &bflib::input_char},
	{"bflib_fetch_char", llvm::Type::IntTy, args_noargs, 0, false, &bflib::fetch_char},
	{"bflib_store_char", llvm::Type::VoidTy, args_1int, 1, false, &bflib::store_char},
	{"bflib_fungespace_get", llvm::Type::IntTy, args_vector, 1, false, &bflib::fungespace_get},
	{"bflib_fungespace_put", llvm::Type::VoidTy, args_vector_int, 2, false, &bflib::fungespace_put},
	{"bflib_go_away", llvm::Type::VoidTy, args_noargs, 0, false, &bflib::go_away},
	{"bflib_clear_stack", llvm::Type::VoidTy, args_noargs, 0, false, &bflib::clear_stack},
	{0,0,0,false,0}
};

llvm::Module *load_bflib()
{
	std::auto_ptr<llvm::Module> mod(llvm::ParseBytecodeBuffer(bflib_bytecode, bflib_bytecode_size));
	
	for (int n = 0; bfuncs[n].name; n++)
	{
		std::vector<const llvm::Type*> args(bfuncs[n].args, bfuncs[n].args + bfuncs[n].nargs);
		*bfuncs[n].holder = mod->getFunction(std::string(bfuncs[n].name), llvm::FunctionType::get(bfuncs[n].ret, args, bfuncs[n].vararg));
		assert(*bfuncs[n].holder && "bflib function not found");
	}

	return mod.release();
}

llvm::Value *get_thread_vector_value(int x1, int x2, llvm::BasicBlock *cb)
{
	assert(x1 >= 0 && x1 < 3);
	llvm::Value *thr = new llvm::LoadInst(bflib::thread, "thr", cb);
	
	std::vector<llvm::Value *> indices;
	// remove pointer
	indices.push_back(llvm::ConstantSInt::get(llvm::Type::IntTy, 0)); 
	// select IP / delta / shift
	indices.push_back(llvm::ConstantUInt::get(llvm::Type::UIntTy, x1)); 
	// remove struct
	indices.push_back(llvm::ConstantUInt::get(llvm::Type::UIntTy, 0)); 
	// select dim
	indices.push_back(llvm::ConstantSInt::get(llvm::Type::IntTy, x2)); 

	llvm::Value *v = new llvm::GetElementPtrInst(thr, indices, "", cb);
	return v;
}

