// Funge JIT compiler
// Copyright (c) 2004 Marek Materzok
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.


#include <iostream>
#include <vector>
#include <cassert>

#include <cstdio>

#include <llvm/ExecutionEngine/ExecutionEngine.h>
#include <llvm/ExecutionEngine/GenericValue.h>
#include <llvm/Bytecode/Reader.h>
#include <llvm/ModuleProvider.h>
#include <llvm/Module.h>
#include <llvm/Type.h>
#include <llvm/DerivedTypes.h>
#include <llvm/Constants.h>
#include <llvm/Instructions.h>
#include <llvm/PassManager.h>
#include <llvm/Transforms/Scalar.h>
#include <llvm/Target/TargetData.h>

#include <boost/scoped_ptr.hpp>
#include <boost/format.hpp>

#include "Bflib.hpp"
#include "BflibInlinePass.hpp"
#include "BflibLowerVectorPass.hpp"
#include "BflibSimplifyStackPass.hpp"

std::vector<llvm::Function *> funs;

llvm::Module *module;
llvm::ExecutionEngine *engine;
llvm::FunctionPassManager *passmgr;

void read_fungespace(std::istream &in)
{
	char bufor[16384];
	Vector vec;
	for (int n = 0; n < DIMENSIONS; n++) vec.v[n] = 0;
	while (in)
	{
		in.getline(bufor, 16384);
		for (int x = 0; bufor[x]; x++)
		{
			vec.v[0] = x;
			Cell *c = bflib_get_cell(&vec);
			c->value = bufor[x];
		}
		vec.v[1]++;
	}
}

typedef void (*voidfun)(void);

void update_thread(Thread &t, llvm::BasicBlock *cb)
{
	for (int x = 0; x < DIMENSIONS; x++)
	{
		if (t.ip.v[x] != bflib_thread->ip.v[x])
		{
			llvm::Value *v = get_thread_vector_value(0, x, cb);
			new llvm::StoreInst(llvm::ConstantSInt::get(llvm::Type::IntTy, t.ip.v[x]), v, cb);
		}
		if (t.delta.v[x] != bflib_thread->delta.v[x])
		{
			llvm::Value *v = get_thread_vector_value(1, x, cb);
			new llvm::StoreInst(llvm::ConstantSInt::get(llvm::Type::IntTy, t.delta.v[x]), v, cb);
		}
	}
}

llvm::Function *generate_code(int id)
{
	llvm::Function *ret = module->getOrInsertFunction((boost::format("fungefun%1%") % id).str(), llvm::Type::VoidTy, 0);
	ret->deleteBody();
	Thread t = *bflib_thread;
	Vector &delta = t.delta;
	Vector &ip = t.ip;
	llvm::BasicBlock *cb = new llvm::BasicBlock("entry", ret);
	llvm::Value *temp_vector = new llvm::AllocaInst(bflib::vector, 0, "tempvector", cb);
	bool gen_cond = true;
	bool stringmode = false;

	// used in instruction code
	llvm::Value *v1, *v2, *v3, *v4;
	llvm::Instruction::BinaryOps bop;
	Thread jt;
	llvm::BasicBlock *nbt, *nbf, *nbm;
	
	while (gen_cond)
	{
//		std::cerr << delta << ip << std::endl;
		Cell *curcell = bflib_get_cell(&ip);
		int tmp = 0;

		if (stringmode)
		{
			if (curcell->value == '"') stringmode = false;
			else
			{
				new llvm::CallInst(bflib::push_int, llvm::ConstantSInt::get(llvm::Type::IntTy, curcell->value), "", cb);
			}
		}
		else
		switch (curcell->value)
		{
			// basic stack ops
			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				new llvm::CallInst(bflib::push_int, llvm::ConstantSInt::get(llvm::Type::IntTy, curcell->value - '0'), "", cb);
				break;
			case 'a': case 'b': case 'c':
			case 'd': case 'e': case 'f':
				new llvm::CallInst(bflib::push_int, llvm::ConstantSInt::get(llvm::Type::IntTy, curcell->value - 'a' + 10), "", cb);
				break;
			case '$':
				new llvm::CallInst(bflib::pop_int, "", cb);
				break;
			case '\\':
				v2 = new llvm::CallInst(bflib::pop_int, "", cb);
				v1 = new llvm::CallInst(bflib::pop_int, "", cb);
				new llvm::CallInst(bflib::push_int, v2, "", cb);
				new llvm::CallInst(bflib::push_int, v1, "", cb);
				break;
			case ':':
				v1 = new llvm::CallInst(bflib::pop_int, "", cb);
				new llvm::CallInst(bflib::push_int, v1, "", cb);
				new llvm::CallInst(bflib::push_int, v1, "", cb);
				break;
			// arithmetic
			case '+': bop = llvm::Instruction::Add; goto binary_label;
			case '-': bop = llvm::Instruction::Sub; goto binary_label;
			case '*': bop = llvm::Instruction::Mul; goto binary_label;
			case '/': bop = llvm::Instruction::Div; goto binary_label;
			case '%': bop = llvm::Instruction::Rem; goto binary_label;
			binary_label:
				v2 = new llvm::CallInst(bflib::pop_int, "", cb);
				v1 = new llvm::CallInst(bflib::pop_int, "", cb);
				v3 = llvm::BinaryOperator::create(bop, v1, v2, "", cb);
				new llvm::CallInst(bflib::push_int, v3, "", cb);
				break;
			case '!':
				v1 = new llvm::CallInst(bflib::pop_int, "", cb);
				v1 = new llvm::CastInst(v1, llvm::Type::BoolTy, "", cb);
				v3 = llvm::BinaryOperator::createNot(v1, "", cb);
				v3 = new llvm::CastInst(v3, llvm::Type::IntTy, "", cb);
				new llvm::CallInst(bflib::push_int, v3, "", cb);
				break;
			case '`':
				v2 = new llvm::CallInst(bflib::pop_int, "", cb);
				v1 = new llvm::CallInst(bflib::pop_int, "", cb);
				v3 = new llvm::SetCondInst(llvm::Instruction::SetGT, v1, v2, "", cb);
				v3 = new llvm::CastInst(v3, llvm::Type::IntTy, "", cb);
				new llvm::CallInst(bflib::push_int, v3, "", cb);
				break;
			// direction change ops
			case '^': delta.v[0] = 0; delta.v[1] = -1; delta.v[2] = 0; break;
			case 'v': delta.v[0] = 0; delta.v[1] = 1; delta.v[2] = 0; break;
			case '<': delta.v[0] = -1; delta.v[1] = 0; delta.v[2] = 0; break;
			case '>': delta.v[0] = 1; delta.v[1] = 0; delta.v[2] = 0; break;
			case 'h': delta.v[0] = 0; delta.v[1] = 0; delta.v[2] = -1; break;
			case 'l': delta.v[0] = 0; delta.v[1] = 0; delta.v[2] = 1; break;
			case 'r':
				delta.v[0] = -delta.v[0];
				delta.v[1] = -delta.v[1];
				delta.v[2] = -delta.v[2];
				break;
			case '[':
				tmp = delta.v[0];
				delta.v[0] = delta.v[1];
				delta.v[1] = -tmp;
				break;
			case ']':
				tmp = delta.v[0];
				delta.v[0] = -delta.v[1];
				delta.v[1] = tmp;
				break;
			case '#':
				bflib_vector_wrapadd(&ip, &delta);
				break;
			case '?':
				update_thread(t, cb);
				new llvm::CallInst(bflib::go_away, "", cb);
				new llvm::ReturnInst(cb);
				gen_cond = false; 
				break;
			// conditionals
			case 'm': tmp++;
			case '|': tmp++;
			case '_':
				v1 = new llvm::CallInst(bflib::pop_int, "", cb);
				v1 = new llvm::CastInst(v1, llvm::Type::BoolTy, "", cb);

				nbt = new llvm::BasicBlock("true", ret);
				jt = t;
				jt.delta.v[0] = jt.delta.v[1] = jt.delta.v[2] = 0;
				jt.delta.v[tmp] = -1;
				bflib_vector_wrapadd(&jt.ip, &jt.delta);
				update_thread(jt, nbt);
				new llvm::ReturnInst(nbt);

				nbf = new llvm::BasicBlock("false", ret);
				jt = t;
				jt.delta.v[0] = jt.delta.v[1] = jt.delta.v[2] = 0;
				jt.delta.v[tmp] = 1;
				bflib_vector_wrapadd(&jt.ip, &jt.delta);
				update_thread(jt, nbf);
				new llvm::ReturnInst(nbf);

				new llvm::BranchInst(nbt, nbf, v1, cb);
				gen_cond = false; 
				break;
			case 'w':
				v2 = new llvm::CallInst(bflib::pop_int, "", cb);
				v1 = new llvm::CallInst(bflib::pop_int, "", cb);

				v3 = new llvm::SetCondInst(llvm::Instruction::SetGT, v1, v2, "", cb);
				
				nbt = new llvm::BasicBlock("gt", ret);
				jt = t;
				tmp = jt.delta.v[0]; // ]
				jt.delta.v[0] = -jt.delta.v[1];
				jt.delta.v[1] = tmp;
				bflib_vector_wrapadd(&jt.ip, &jt.delta);
				update_thread(jt, nbt);
				new llvm::ReturnInst(nbt);

				nbm = new llvm::BasicBlock("le", ret);
				v4 = new llvm::SetCondInst(llvm::Instruction::SetLT, v1, v2, "", nbm);

				new llvm::BranchInst(nbt, nbm, v3, cb);

				nbt = new llvm::BasicBlock("lt", ret);
				jt = t;
				tmp = jt.delta.v[0]; // [
				jt.delta.v[0] = jt.delta.v[1];
				jt.delta.v[1] = -tmp;
				bflib_vector_wrapadd(&jt.ip, &jt.delta);
				update_thread(jt, nbt);
				new llvm::ReturnInst(nbt);
				
				nbf = new llvm::BasicBlock("eq", ret);
				jt = t;
				bflib_vector_wrapadd(&jt.ip, &jt.delta);
				update_thread(jt, nbf);
				new llvm::ReturnInst(nbf);
				
				new llvm::BranchInst(nbt, nbf, v4, nbm);

				gen_cond = false;
				break;
			// misc ops
			case '@': 
				gen_cond = false; 
				update_thread(t, cb);
				new llvm::ReturnInst(cb);
				break;
			default: case 'z': break;
			// input / output
			case '&':
				v1 = new llvm::CallInst(bflib::input_int, "", cb);
				new llvm::CallInst(bflib::push_int, v1, "", cb);
				break;
			case '~':
				v1 = new llvm::CallInst(bflib::input_char, "", cb);
				new llvm::CallInst(bflib::push_int, v1, "", cb);
				break;
			case '.':
				v1 = new llvm::CallInst(bflib::pop_int, "", cb);
				new llvm::CallInst(bflib::output_int, v1, "", cb);
				break;
			case ',':
				v1 = new llvm::CallInst(bflib::pop_int, "", cb);
				new llvm::CallInst(bflib::output_char, v1, "", cb);
				break;
			// fungespace manipulation
			case 'g':
				new llvm::CallInst(bflib::pop_vector, temp_vector, "", cb);
				v1 = new llvm::CallInst(bflib::fungespace_get, temp_vector, "", cb);
				new llvm::CallInst(bflib::push_int, v1, "", cb);
				break;
			case 'p':
				new llvm::CallInst(bflib::pop_vector, temp_vector, "", cb);
				v1 = new llvm::CallInst(bflib::pop_int, "", cb);
				new llvm::CallInst(bflib::fungespace_put, temp_vector, v1, "", cb);
				break;
			// stackstack manipulation
			case '{':
				update_thread(t, cb);
				new llvm::CallInst(bflib::begin_block, "", cb);
				new llvm::ReturnInst(cb);
				gen_cond = false; 
				break;
			case '}':
				update_thread(t, cb);
				new llvm::CallInst(bflib::end_block, "", cb);
				new llvm::ReturnInst(cb);
				gen_cond = false; 
				break;
			case 'u':
				update_thread(t, cb);
				new llvm::CallInst(bflib::stack_transfer, "", cb);
				new llvm::ReturnInst(cb);
				gen_cond = false; 
				break;
			// stringmode on
			case '"':
				stringmode = true;
				break;
			case '\'':
				v1 = new llvm::CallInst(bflib::fetch_char, "", cb);
				new llvm::CallInst(bflib::push_int, v1, "", cb);
				bflib_vector_wrapadd(&ip, &delta);
				break;
			case 's':
				v1 = new llvm::CallInst(bflib::pop_int, "", cb);
				new llvm::CallInst(bflib::store_char, v1, "", cb);
				break;
			case 'n':
				new llvm::CallInst(bflib::clear_stack, "", cb);
				break;
			// TODO: unimplemented
			case '(': // Load Semantics
			case ')': // Store Semantics
			case ';': // Jump Over  
			case '=': // Execute (filesystem)
			case 'i': // Input File (filesystem)
			case 'j': // Jump Forward
			case 'k': // Iterate
			case 'o': // Output File (filesystem)
			case 'q': // Quit
			case 't': // Split (threads)
			case 'x': // Absolute Delta
			case 'y': // SysInfo
				std::cerr << "Unknown insn: " << curcell->value << std::endl;
				assert(0 && "unimplemented insn");
		}
		if (gen_cond) bflib_vector_wrapadd(&ip, &delta);
	}
	return ret;
}

void main_loop()
{
	while (true)
	{
		llvm::GenericValue gv = engine->runFunction(bflib::main, std::vector<llvm::GenericValue>());
		int result = gv.IntVal;
		Cell *curcell = bflib_get_cell(&bflib_thread->ip);
		if (curcell->value == '@') break;
		if (result == -1)
		{
			int id = funs.size();
			llvm::Function *f = generate_code(id);
			passmgr->run(*f);
//			std::cerr << *f << std::endl; 
			funs.push_back(f);
			bflib_functions[id] = (voidfun)engine->getPointerToFunction(f);
			
			FunStart *fs = new FunStart();
			fs->id = id;
			fs->delta = bflib_thread->delta;
			fs->next = curcell->fstart;
			curcell->fstart = fs;
		}
	}
}

int main(int argc, char **argv)
{
	module = load_bflib();
	llvm::ModuleProvider *provider = new llvm::ExistingModuleProvider(module);
	engine = llvm::ExecutionEngine::create(provider, false, 0);
	
	passmgr = new llvm::FunctionPassManager(provider);
	passmgr->add(new llvm::TargetData("llfunge", module));
	passmgr->add(new BflibLowerVectorPass());
	passmgr->add(new BflibSimplifyStackPass());
	passmgr->add(new BflibInlinePass());
	passmgr->add(llvm::createReassociatePass());
	passmgr->add(llvm::createGCSEPass());
	passmgr->add(llvm::createInstructionCombiningPass());
	passmgr->add(llvm::createCFGSimplificationPass());
	passmgr->add(llvm::createDeadCodeEliminationPass());

	engine->runFunction(bflib::init, std::vector<llvm::GenericValue>());

	for (llvm::Module::giterator iter = module->gbegin(); iter != module->gend(); ++iter)
	{
		if (iter->getName() == "bflib_stackstack")
			bflib_stackstack = static_cast<Stack*>(*static_cast<void**>(
						engine->getOrEmitGlobalVariable(&*iter)));
		if (iter->getName() == "bflib_topstack")
			bflib_topstack = static_cast<Stack*>(*static_cast<void**>(
						engine->getOrEmitGlobalVariable(&*iter)));
		if (iter->getName() == "bflib_thread")
		{
			bflib_thread = static_cast<Thread*>(*static_cast<void**>(
						engine->getOrEmitGlobalVariable(&*iter)));
			bflib::thread = &*iter;
		}
		if (iter->getName() == "bflib_fspace")
			bflib_fspace = static_cast<FungeSpace*>(*static_cast<void**>(
						engine->getOrEmitGlobalVariable(&*iter)));
		if (iter->getName() == "bflib_functions")
			bflib_functions = static_cast<void (**)(void)>(*static_cast<void**>(
						engine->getOrEmitGlobalVariable(&*iter)));
	}

	// TODO: czytanie z pliku
	read_fungespace(std::cin);
	
	bflib_fspace->ndims = 2;

	main_loop();
}

